# ruff: noqa: D100, D101, D106
# mypy: disable-error-code="call-arg,valid-type,name-defined,type-arg"
from datetime import datetime

from pydantic import BaseModel, ConfigDict, Field


class BaseModelWithTimestamp(BaseModel):
    created_at: datetime = Field(
        ...,
        examples=[datetime.now()],
        description="ISO-8601 timestamp",
    )
    updated_at: datetime = Field(
        ...,
        examples=[datetime.now()],
        description="ISO-8601 timestamp",
    )

    model_config = ConfigDict(from_attributes=True)



# ruff: noqa: D101

"""Кастомные ошибки для handler и responses."""

from http import HTTPStatus

from axiom_core.exception.exception import CustomHTTPException


class InternalServerException(CustomHTTPException):
    def __init__(self, detail: str = "Internal server error"):
        super().__init__(status_code=HTTPStatus.INTERNAL_SERVER_ERROR, detail=detail)


class BadRequestException(CustomHTTPException):
    def __init__(self, detail: str = "Bad Request"):
        super().__init__(status_code=HTTPStatus.BAD_REQUEST, detail=detail)


class NotFoundException(CustomHTTPException):
    def __init__(self, detail: str = "Not found"):
        super().__init__(status_code=HTTPStatus.NOT_FOUND, detail=detail)


class UnprocessableEntityException(CustomHTTPException):
    def __init__(self, detail: str = "Unprocessable entity"):
        super().__init__(status_code=HTTPStatus.UNPROCESSABLE_ENTITY, detail=detail)


class PayloadTooLargeException(CustomHTTPException):
    def __init__(self, message: str = "Payload too large"):
        super().__init__(
            status_code=HTTPStatus.REQUEST_ENTITY_TOO_LARGE,
            detail=message,
        )


class UnsupportedMediaTypeException(CustomHTTPException):
    def __init__(self, message: str = "Unsupported media type"):
        super().__init__(status_code=HTTPStatus.UNSUPPORTED_MEDIA_TYPE, detail=message)


class TooEarlyException(CustomHTTPException):
    def __init__(self, message: str = "Too early"):
        super().__init__(status_code=HTTPStatus.TOO_EARLY, detail=message)


class FieldException(CustomHTTPException):
    def __init__(self, detail: str = "Unacceptable field."):
        super().__init__(status_code=HTTPStatus.UNPROCESSABLE_ENTITY, detail=detail)



# ruff: noqa: W505
# mypy: disable-error-code="name-defined,valid-type"
"""Базовый репозиторий."""

from abc import ABC, abstractmethod
from collections import defaultdict
from enum import Enum
from typing import Any

from axiom_core.dto.scheme.request.filter import FilterParam, FilterRequest
from axiom_core.dto.type.filter import FilterType
from axiom_core.dto.type.query import QueryOperator
from axiom_core.dto.type.sort import SortTypeEnum
from axiom_core.exception import BadRequestException, FieldException


class BaseRepository[ModelType, SessionType, QueryType](ABC):
    """Базовый класс для репозиториев данных."""

    def __init__(self, model: ModelType, db_session: SessionType):
        self.session = db_session
        self.model_class = model

    @abstractmethod
    async def create(self, attributes: dict[str, Any]) -> ModelType:
        """Создает экземпляр модели.

        Args:
            attributes: Атрибуты для создания модели.

        Returns:
            Созданный экземпляр модели.
        """
        raise NotImplementedError

    @abstractmethod
    async def create_model(self, model: ModelType) -> ModelType:
        """Создает экземпляр модели.

        Args:
            model: Модель базы данных.

        Returns:
            Созданный экземпляр модели.
        """
        raise NotImplementedError

    @abstractmethod
    async def create_many(
        self,
        attributes_list: list[dict[str, Any]],
    ) -> list[ModelType]:
        """Создает несколько экземпляров модели за раз."""
        raise NotImplementedError

    async def get_all(
        self,
        skip: int = 0,
        limit: int = 100,
        sort_by: str | None = None,
        sort_type: SortTypeEnum | None = SortTypeEnum.asc,
    ) -> list[ModelType]:
        """Возвращает список экземпляров модели.

        Args:
            skip: Количество записей для пропуска.
            limit: Количество записей для возврата.
            sort_by: Поле для сортировки.
            sort_type: Направление сортировки.

        Returns:
            Список экземпляров модели.
        """
        query = self._query()
        query = self._sort_by(query=query, sort_by=sort_by, sort_type=sort_type)
        query = self._paginate(query=query, skip=skip, limit=limit)

        return await self._all(query)

    async def get_by(
        self,
        field: str,
        value: Any,
        operator: QueryOperator = QueryOperator.EQUALS,
        skip: int = 0,
        limit: int = 100,
        sort_by: str | None = None,
        sort_type: SortTypeEnum | None = SortTypeEnum.asc,
        unique: bool = False,
    ) -> ModelType | None | list[ModelType]:
        """Возвращает экземпляр модели, соответствующий полю и значению.

        Args:
            field: Поле для совпадения.
            value: Значение для совпадения.
            operator: Оператор сравнения.
            skip: Количество записей для пропуска.
            limit: Количество записей для возврата.
            sort_by: Поле для сортировки.
            sort_type: Направление сортировки.
            unique: Уникальный параметр.

        Returns:
            Экземпляр модели.
        """
        query = self._query()
        query = self._maybe_join(query=query, field=field)
        query = self._filter(
            query=query,
            filter_request=FilterRequest(
                filters=[FilterParam(field=field, value=value, operator=operator)],
                type=FilterType.OR,
            ),
        )

        if unique:
            return await self._one_or_none(query)

        query = self._sort_by(query=query, sort_by=sort_by, sort_type=sort_type)
        query = self._paginate(query=query, skip=skip, limit=limit)

        return await self._all(query)

    async def get_by_filters(
        self,
        filter_request: FilterRequest | None = None,
        skip: int = 0,
        limit: int = 100,
        sort_by: str | None = None,
        sort_type: SortTypeEnum | None = SortTypeEnum.asc,
        unique: bool = False,
    ) -> ModelType | None | list[ModelType]:
        """Возвращает экземпляры модели, соответствующие фильтрам.

        Args:
            filter_request: Фильтры для совпадения.
            skip: Количество записей для пропуска.
            limit: Количество записей для возврата.
            sort_by: Поле для сортировки.
            sort_type: Направление сортировки.
            unique: Уникальная запись.

        Returns:
            Список экземпляров модели.
        """
        query = self._query()
        if filter_request is not None and len(filter_request.filters) > 0:
            for param in filter_request.filters:
                query = self._maybe_join(query=query, field=param.field)

            query = self._filter(query=query, filter_request=filter_request)

        data: ModelType | None | list[ModelType]

        if unique:
            data = await self._one_or_none(query)
        else:
            query = self._sort_by(query=query, sort_by=sort_by, sort_type=sort_type)
            query = self._paginate(query=query, skip=skip, limit=limit)

            data = await self._all(query)

        return data

    async def count(self, filter_request: FilterRequest | None = None) -> int:
        """Возвращает кол-во записей, соответствующих фильтрам.

        Args:
            filter_request: Фильтры для совпадения.

        Returns:
            Кол-во записей.
        """
        query = self._query()
        if filter_request is not None and len(filter_request.filters) > 0:
            for param in filter_request.filters:
                query = self._maybe_join(query=query, field=param.field)
            query = self._filter(query, filter_request)
        return await self._count(query)

    @abstractmethod
    async def update(self, model: ModelType, attributes: dict[str, Any]) -> ModelType:
        """Обновляет экземпляр модели с заданными атрибутами.

        Args:
            model: Модель для обновления.
            attributes: Атрибуты для обновления экземпляра модели.

        Returns:
            Обновленный экземпляр модели.
        """
        raise NotImplementedError

    @abstractmethod
    async def update_by(
        self,
        field: str,
        value: Any,
        attributes: dict[str, Any],
        operator: QueryOperator = QueryOperator.EQUALS,
        unique: bool = False,
    ) -> ModelType | None | list[ModelType]:
        """Обновляет сущности, найденные по заданному условию, за один запрос.

        Args:
            field: Поле для поиска.
            value: Значение для поиска.
            attributes: Словарь {поле: значение} для обновления.
            operator: Оператор сравнения (по умолчанию EQUALS).
            unique: Флаг, что ожидается ровно одна запись (или ни одной).

        Returns:
            - Если unique=True, возвращается одна обновлённая сущность или None,
              если ничего не найдено.
            - Если unique=False, возвращается список всех обновлённых сущностей
              (включая пустой список, если ничего не найдено).
        """
        raise NotImplementedError

    @abstractmethod
    async def update_by_filters(
        self,
        filter_request: FilterRequest,
        attributes: dict[str, Any],
        unique: bool = False,
    ) -> ModelType | None | list[ModelType]:
        """Обновляет экземпляры модели по фильтрам.

        Args:
           filter_request: Фильтры для поиска.
           attributes: Атрибуты для обновления экземпляров модели.
           unique: Флаг, что ожидается ровно одна запись (или ни одной).

        Returns:
            Обновлённые экземпляры модели.
        """
        raise NotImplementedError

    @abstractmethod
    async def delete(self, model: ModelType) -> ModelType:
        """Удаляет экземпляр модели.

        Args:
            model: Модель для удаления.

        Returns:
            Удаленный экземпляр модели.
        """
        raise NotImplementedError

    @abstractmethod
    async def delete_by(
        self,
        field: str,
        value: Any,
        operator: QueryOperator = QueryOperator.EQUALS,
        unique: bool = False,
    ) -> ModelType | None | list[ModelType]:
        """Удаляет сущности, найденные по заданному условию, за один запрос.

        Args:
            field: Поле для поиска.
            value: Значение для поиска.
            operator: Оператор сравнения (по умолчанию EQUALS).
            unique: Флаг, что ожидается ровно одна запись (или ни одной).

        Returns:
            - Если unique=True, возвращается одна удаленная сущность или None,
              если ничего не найдено.
            - Если unique=False, возвращается список всех удаленных сущностей
              (включая пустой список, если ничего не найдено).
        """
        raise NotImplementedError

    @abstractmethod
    async def delete_by_filters(
        self,
        filter_request: FilterRequest,
        unique: bool = False,
    ) -> ModelType | None | list[ModelType]:
        """Удаляет экземпляры модели по фильтрам.

        Args:
           filter_request: Фильтры для поиска.
           unique: Флаг, что ожидается ровно одна запись (или ни одной).

        Returns:
            Удалённые экземпляры модели.
        """
        raise NotImplementedError

    @abstractmethod
    async def create_or_update_by(
        self,
        attributes: dict[str, Any],
        update_fields: list[str] | None = None,
    ) -> ModelType:
        """Создать или обновить модель."""
        raise NotImplementedError

    @abstractmethod
    def _query(self) -> QueryType:
        """Возвращает вызываемый объект, который можно использовать для запроса модели.

        Returns:
            Вызываемый объект, который можно использовать для запроса модели.
        """
        raise NotImplementedError

    @abstractmethod
    def _maybe_join(self, query: QueryType, field: str) -> QueryType:
        """Возвращает запрос, который может указать на использование связанной сущности.

        Returns:
            Запрос со связанной сущностью.
        """
        raise NotImplementedError

    @abstractmethod
    async def _all(self, query: QueryType) -> list[ModelType]:
        """Возвращает все результаты запроса.

        Args:
            query: Запрос для выполнения.

        Returns:
            Список экземпляров модели.
        """
        raise NotImplementedError

    @abstractmethod
    async def _one_or_none(self, query: QueryType) -> ModelType | None:
        """Возвращает первый результат запроса или None.

        Args:
            query: Запрос для выполнения.

        Returns:
            Экземпляр модели.
        """
        raise NotImplementedError

    @abstractmethod
    def _get_by[ExpressionType](
        self,
        field: str,
        value: Any,
        operator: QueryOperator = QueryOperator.EQUALS,
    ) -> ExpressionType:  # type: ignore[type-var]
        """Возвращает запрос по указанному полю.

        Args:
            field: Колонка для фильтрации.
            value: Значение для фильтрации.
            operator: Оператор сравнения.

        Notes:
            В этом месте происходит валидация поля и значения.

        Returns:
            Выражение для нужного оператора.
        """
        raise NotImplementedError

    @abstractmethod
    def _filter(self, query: QueryType, filter_request: FilterRequest) -> QueryType:
        """Добавляет фильтры для query.

        Args:
            query: Запрос.
            filter_request: Запрос для фильтров.

        Notes:
            Использует _get_by для получения значения поля.

        Returns:
            Запрос(QueryType)
        """
        raise NotImplementedError

    @abstractmethod
    async def _count(self, query: QueryType) -> int:
        """Возвращает количество записей.

        Args:
            query: Запрос для выполнения.

        Returns:
            Количество экземпляров модели.
        """
        raise NotImplementedError

    @abstractmethod
    def _paginate(self, query: QueryType, skip: int = 0, limit: int = -1) -> QueryType:
        """Возвращает запрос, в котором применена пагинация.

        Args:
            query: Запрос для сортировки.
            skip: Количество записей для пропуска.
            limit: Количество записей для возврата.

        Returns:
            Отпагиннированный запрос.
        """
        raise NotImplementedError

    @abstractmethod
    def _sort_by(
        self,
        query: QueryType,
        sort_by: str | None,
        sort_type: SortTypeEnum | None = SortTypeEnum.asc,
    ) -> QueryType:
        """Возвращает запрос, отсортированный по указанной колонке.

        Args:
            query: Запрос для сортировки.
            sort_by: Поле для сортировки.
            sort_type: Направление сортировки.

        Returns:
            Отсортированный запрос.
        """
        raise NotImplementedError

    def _get_deep_unique_from_dict(
        self,
        columns: dict[str, Any],
    ) -> dict[str, Any] | list[dict[str, Any]]:
        """Рекурсивно обходит словарь/список/примитив.

        1) Если встретили словарь - углубляемся по ключам.
        2) Если встретили список из словарей - собираем уникальные значения по
            ключам, рекурсивно обрабатывая значения.
        3) Если встретили список "простых" (или смешанных) элементов - рекурсивно
            обрабатываем каждый элемент и сохраняем только уникальные.
        4) Если встретили примитив (int, str, bool и т.д.) - возвращаем как есть.
        """
        # Если это словарь, обрабатываем по ключам
        if isinstance(columns, dict):
            result = {}
            for key, value in columns.items():
                result[key] = self._get_deep_unique_from_dict(value)
            return result

        # Если это список
        elif isinstance(columns, list):
            # Проверяем, что все элементы - словари
            if all(isinstance(item, dict) for item in columns):
                # Собираем уникальные значения по ключам
                aggregated = defaultdict(list)
                for item in columns:
                    for k, v in item.items():
                        processed_v = self._get_deep_unique_from_dict(v)
                        if processed_v not in aggregated[k]:
                            aggregated[k].append(processed_v)
                return self._get_deep_unique_from_dict(aggregated)

            else:
                # Список не из одних dict (или смешанный) — рекурсивно обрабатываем каждый элемент
                processed_list = []
                for item in columns:
                    processed_item = self._get_deep_unique_from_dict(item)
                    if processed_item not in processed_list:
                        processed_list.append(processed_item)
                return processed_list

        # Если это не список и не словарь (примитив: int, str, bool, и т.п.)
        else:
            return columns

    @staticmethod
    def _check_field_relation_depth(field: str, depth: int = 1) -> None:
        if field.count(".") > depth:
            raise BadRequestException(f"Supported relation depth no more than {depth}")

    def _get_model_field_type(self, _model: ModelType, _field: str) -> type:
        """Получить python-style тип поля."""
        raise NotImplementedError

    def _resolve_field_relation(self, field: str) -> tuple[ModelType, str]:
        """Получение модели и имени колонки из поля."""
        raise NotImplementedError

    def _validate_params(self, field: str, value: Any | None = None) -> None:
        """Валидация параметров по наличию поля в модели и типу значения."""
        model, column_name = self._resolve_field_relation(field)

        model_field_type = self._get_model_field_type(model, column_name)
        if issubclass(model_field_type, dict):
            return None
        if issubclass(model_field_type, Enum):
            if all(value != item for item in model_field_type):
                raise FieldException(
                    f"Value {value} is not permissible for the field {field}. "
                    f"Available values: {[e for e in model_field_type]}",
                )
            enum_member = next(iter(model_field_type)).value
            model_field_type = type(enum_member)
        if value is not None and not isinstance(value, model_field_type):
            raise FieldException(
                f"Wrong type for field {field}: "
                f"expected {model_field_type.__name__}, "
                f"received {type(value).__name__}",
            )
        return None



# ruff: noqa: D100, D101, D106
# mypy: disable-error-code="call-arg,valid-type,name-defined,type-arg"

from pydantic import BaseModel, ConfigDict, Field


class PaginationResponse[T](BaseModel):
    page: int = Field(..., examples=[0])
    page_size: int = Field(..., examples=[0])
    total_pages: int = Field(..., examples=[0])
    total_count: int = Field(..., examples=[0])
    data: list[T] = Field(...)

    model_config = ConfigDict(from_attributes=True, arbitrary_types_allowed=True)



# ruff: noqa: D100

from enum import StrEnum


class ModeType(StrEnum):
    """Место запуска проекта локальное/тестовое в докере/продакшн."""

    local = "local"
    dev = "dev"
    production = "production"



# ruff: noqa: D100, D101

from enum import StrEnum


class QueryOperator(StrEnum):
    EQUALS = "EQUALS"
    NOT_EQUAL = "NOT_EQUAL"
    IN = "IN"
    NOT_IN = "NOT_IN"
    GREATER = "GREATER"
    EQUALS_OR_GREATER = "EQUALS_OR_GREATER"
    LESS = "LESS"
    EQUALS_OR_LESS = "EQUALS_OR_LESS"
    STARTS_WITH = "STARTS_WITH"
    NOT_START_WITH = "NOT_START_WITH"
    ENDS_WITH = "ENDS_WITH"
    NOT_END_WITH = "NOT_END_WITH"
    CONTAINS = "CONTAINS"
    NOT_CONTAIN = "NOT_CONTAIN"



"""Модуль кастомных exception."""

from .base import (
    BadRequestException,
    FieldException,
    InternalServerException,
    NotFoundException,
    PayloadTooLargeException,
    TooEarlyException,
    UnprocessableEntityException,
    UnsupportedMediaTypeException,
)

__all__ = [
    "InternalServerException",
    "BadRequestException",
    "NotFoundException",
    "UnprocessableEntityException",
    "PayloadTooLargeException",
    "TooEarlyException",
    "UnsupportedMediaTypeException",
    "FieldException",
]



# ruff: noqa: D100, D101, D102
from typing import Any

from pydantic import BaseModel, Field

from axiom_core.dto.type.filter import FilterType
from axiom_core.dto.type.query import QueryOperator


class FilterParam(BaseModel):
    field: str = Field(..., examples=["message"])
    value: Any = Field(..., examples=["abc"])
    operator: QueryOperator = Field(..., examples=[QueryOperator.EQUALS])


class FilterRequest(BaseModel):
    filters: list[FilterParam] = Field([])
    type: FilterType = Field(FilterType.OR)

    def __repr__(self) -> str:
        """Представление данных схемы для ключа."""
        filters_repr = (
            "; ".join(f"{filter_param}" for filter_param in self.filters)
            if self.filters
            else "None"
        )
        return f"{self.__class__.__name__}(filters=[{filters_repr}], type={self.type})"



# ruff: noqa: D100, D101
from enum import StrEnum

from pydantic import BaseModel, Field


class SortTypeEnum(StrEnum):
    asc = "asc"
    desc = "desc"


class SortParams(BaseModel):
    sort_by: str | None = Field(None, examples=["id"])
    sort_type: SortTypeEnum | None = Field(None, examples=["asc", "desc"])



# ruff: noqa: D100

from http.client import HTTPException
from typing import Any

from axiom_core.dto.scheme.response.error import ErrorResponse


class CustomHTTPException(HTTPException):
    """Родитель кастомного HTTP исключения."""

    def __init__(self, status_code: int, detail: str = "Server error"):
        self.status_code = status_code
        self.detail = detail

    def as_dict(self) -> dict[str, Any]:
        """Представление ошибки как dict."""
        return ErrorResponse(
            error_code=self.status_code,
            detail=self.detail,
        ).model_dump()

    def as_response(self) -> ErrorResponse:
        """Представление ошибки как ErrorResponse схема."""
        return ErrorResponse(error_code=self.status_code, detail=self.detail)



# ruff: noqa: W505
# mypy: disable-error-code="type-arg,valid-type,name-defined,arg-type,attr-defined"
"""Базовый контроллер."""

from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from typing import Any
from uuid import UUID

from pydantic import BaseModel

from axiom_core.dto.scheme.request.filter import FilterParam, FilterRequest
from axiom_core.dto.scheme.response.count import CountResponse
from axiom_core.dto.scheme.response.pagination import PaginationResponse
from axiom_core.dto.type.filter import FilterType
from axiom_core.dto.type.query import QueryOperator
from axiom_core.dto.type.sort import SortTypeEnum
from axiom_core.exception import NotFoundException, UnprocessableEntityException
from axiom_core.repository import BaseRepository


class BaseController[ModelType](ABC):
    """Базовый класс для контроллера данных."""

    def __init__(
        self,
        model: ModelType,
        repository: BaseRepository,
        exclude_fields: list[str],
    ):
        self.model_class = model
        self.repository = repository
        self.exclude_fields = exclude_fields

    @abstractmethod
    async def processing_transaction(
        self,
        function: Callable,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """Метод для обработки транзакции."""
        raise NotImplementedError

    def transactional(function):  # type: ignore
        """Декоратор для транзакций."""

        async def wrapper(self, *args, **kwargs):  # type: ignore
            """Функция-обертка."""
            response = await self.processing_transaction(function, *args, **kwargs)
            return response

        return wrapper

    async def get_by(
        self,
        field: str,
        value: Any,
        operator: QueryOperator = QueryOperator.EQUALS,
        skip: int = 0,
        limit: int = 100,
        sort_by: str | None = None,
        sort_type: SortTypeEnum | None = SortTypeEnum.asc,
        unique: bool = False,
    ) -> ModelType | PaginationResponse:
        """Возвращает экземпляр модели, соответствующий значению.

        Args:
            field: Поле для получения.
            value: Значение для совпадения.
            operator: Оператор сравнения.
            skip: Количество записей для пропуска.
            limit: Количество записей для возврата.
            sort_by: Поле для сортировки.
            sort_type: Направление сортировки.
            unique: Уникальность значения.

        Returns:
            Экземпляры модели.
        """
        db_obj = await self.repository.get_by(
            field=field,
            value=value,
            skip=skip,
            limit=limit,
            sort_by=sort_by,
            sort_type=sort_type,
            operator=operator,
            unique=unique,
        )
        if not db_obj:
            raise NotFoundException(
                f"{self.model_class.__name__} {field} with value {value} not exist",  # noqa: E501
            )

        if unique:
            return db_obj  # type: ignore[return-value]

        return await self.make_pagination_response(
            data=db_obj,
            skip=skip,
            limit=limit,
            filter_request=FilterRequest(
                filters=[FilterParam(field=field, value=value, operator=operator)],
                type=FilterType.AND,
            ),
        )

    async def get_by_filters(
        self,
        filter_request: FilterRequest | None = None,
        skip: int = 0,
        limit: int = 100,
        sort_by: str | None = None,
        sort_type: SortTypeEnum | None = SortTypeEnum.asc,
        unique: bool = False,
    ) -> ModelType | None | PaginationResponse:
        """Получает экземпляр модели, соответствующий фильтрам.

        Args:
            filter_request: Фильтры для совпадения.
            skip: Количество записей для пропуска.
            limit: Количество записей для возврата.
            sort_by: Поле для сортировки.
            sort_type: Направление сортировки.
            unique: Уникальность значения.

        Returns:
            Экземпляры модели или уникальный экземпляр.
        """
        models = await self.repository.get_by_filters(
            filter_request=filter_request,
            skip=skip,
            limit=limit,
            sort_by=sort_by,
            sort_type=sort_type,
            unique=unique,
        )

        if unique:
            if models:
                return models  # type: ignore[return-value]
            else:
                raise NotFoundException(
                    f"Unique {self.model_class.__name__} with provided filters not exist",
                )

        return await self.make_pagination_response(
            data=models,
            skip=skip,
            limit=limit,
            filter_request=filter_request,
        )

    async def count(self, filter_request: FilterRequest | None = None) -> CountResponse:
        """Возвращает количество записей в модели по фильтрам.

        Args:
            filter_request: Фильтры для совпадения.

        Returns:
            Количество записей.
        """
        return CountResponse(
            count=await self.repository.count(filter_request=filter_request),
        )

    async def get_by_id(self, id_: int) -> ModelType:
        """Возвращает экземпляр модели, соответствующий идентификатору.

        Args:
            id_: Идентификатор для совпадения.

        Returns:
            Экземпляр модели.
        """
        db_obj = await self.repository.get_by(field="id", value=id_, unique=True)
        if not db_obj:
            raise NotFoundException(
                f"{self.model_class.__name__} with id: {id_} not found",
            )

        return db_obj  # type: ignore[return-value]

    async def get_by_uuid(self, uuid: UUID) -> ModelType:
        """Возвращает экземпляр модели, соответствующий UUID.

        Args:
            uuid: UUID для совпадения.

        Returns:
            Экземпляр модели.
        """
        db_obj = await self.repository.get_by(field="uuid", value=uuid, unique=True)
        if not db_obj:
            raise NotFoundException(
                f"{self.model_class.__name__} with uuid: {uuid} not found",
            )
        return db_obj  # type: ignore[return-value]

    async def get_all(
        self,
        skip: int = 0,
        limit: int = 100,
        sort_by: str | None = None,
        sort_type: SortTypeEnum | None = SortTypeEnum.asc,
    ) -> PaginationResponse:
        """Возвращает список записей на основе параметров пагинации.

        Args:
            skip: Количество записей для пропуска.
            limit: Количество записей для возврата.
            sort_by: Поле для сортировки.
            sort_type: Направление сортировки.

        Returns:
            Список записей.
        """
        response = await self.repository.get_all(
            skip=skip,
            limit=limit,
            sort_by=sort_by,
            sort_type=sort_type,
        )

        return await self.make_pagination_response(
            data=response,
            skip=skip,
            limit=limit,
        )

    @transactional
    async def create(self, attributes: dict[str, Any]) -> ModelType:
        """Создает новый объект в базе данных.

        Args:
            attributes: Атрибуты для создания объекта.

        Returns:
            Созданный объект.
        """
        created_model = await self.repository.create(attributes=attributes)
        return created_model

    @transactional
    async def create_model(self, model: ModelType) -> ModelType:
        """Создает новый объект в базе данных.

        Args:
            model: Модель базы данных

        Returns:
            Созданный объект.
        """
        created_model = await self.repository.create_model(model=model)
        return created_model

    @transactional
    async def create_many(
        self,
        attributes_list: list[dict[str, Any]],
    ) -> list[ModelType]:
        """Создать несколько записей за один запрос."""
        created = await self.repository.create_many(attributes_list)
        return created

    @transactional
    async def delete(self, model: ModelType) -> ModelType:
        """Удаляет объект из базы данных.

        Args:
            model: Модель для удаления.

        Returns:
            Удаленный объект.
        """
        deleted_model = await self.repository.delete(model=model)
        return deleted_model

    @transactional
    async def update(self, model: ModelType, attributes: dict[str, Any]) -> ModelType:
        """Обновляет объект в базе данных.

        Args:
            model: Модель для обновления.
            attributes: Атрибуты для обновления объекта.

        Returns:
            Обновленный объект.
        """
        for field in attributes:
            if field in self.exclude_fields:
                raise UnprocessableEntityException(
                    f"Field {field} is prohibited for updating",
                )
        updated_model = await self.repository.update(model=model, attributes=attributes)
        return updated_model

    @transactional
    async def update_by_filters(
        self,
        filter_request: FilterRequest,
        attributes: dict[str, Any],
        unique: bool = False,
    ) -> ModelType | None | list[ModelType]:
        """Обновляет объекты по филтьрам.

        Args:
            filter_request: Фильтры для поиска.
            attributes: Атрибуты для обновления.
            unique: Если True — ожидается ровно одна запись (или ни одной).

        Returns:
            - Один объект или None, если unique=True.
            - Список объектов, если unique=False.
        """
        for field in attributes:
            if field in self.exclude_fields:
                raise UnprocessableEntityException(
                    f"Field {field} is prohibited for updating",
                )

        result = await self.repository.update_by_filters(
            filter_request=filter_request,
            attributes=attributes,
            unique=unique,
        )

        if unique and result is None:
            raise NotFoundException(
                f"Unique {self.model_class.__name__} with provided filters not exist",
            )

        return result

    @transactional
    async def delete_by_filters(
        self,
        filter_request: FilterRequest,
        unique: bool = False,
    ) -> ModelType | None | list[ModelType]:
        """Удаляет объекты по фильтрам.

        Args:
            filter_request: Фильтры для поиска.
            unique: Если True — ожидается ровно одна запись (или ни одной).

        Returns:
            - Один объект или None, если unique=True.
            - Список объектов, если unique=False.
        """
        result = await self.repository.delete_by_filters(
            filter_request=filter_request,
            unique=unique,
        )

        if unique and result is None:
            raise NotFoundException(
                f"Unique {self.model_class.__name__} with provided filters not exist",
            )

        return result

    @transactional
    async def update_by(
        self,
        field: str,
        value: Any,
        attributes: dict[str, Any],
        operator: QueryOperator = QueryOperator.EQUALS,
        unique: bool = False,
    ) -> ModelType | None | list[ModelType]:
        """Обновляет сущности по одному полю."""
        for f in attributes:
            if f in self.exclude_fields:
                raise UnprocessableEntityException(
                    f"Field {f} is prohibited for updating",
                )

        result = await self.repository.update_by(
            field=field,
            value=value,
            attributes=attributes,
            operator=operator,
            unique=unique,
        )

        if unique and result is None:
            raise NotFoundException(
                f"{self.model_class.__name__} {field} with value {value} not exist",
            )
        return result

    @transactional
    async def delete_by(
        self,
        field: str,
        value: Any,
        operator: QueryOperator = QueryOperator.EQUALS,
        unique: bool = False,
    ) -> ModelType | None | list[ModelType]:
        """Удаляет сущности по одному полю."""
        result = await self.repository.delete_by(
            field=field,
            value=value,
            operator=operator,
            unique=unique,
        )

        if unique and result is None:
            raise NotFoundException(
                f"{self.model_class.__name__} {field} with value {value} not exist",
            )
        return result

    @transactional
    async def update_by_id(self, id_: int, attributes: dict[str, Any]) -> ModelType:
        """Обновляет единственную сущность по её id."""
        for f in attributes:
            if f in self.exclude_fields:
                raise UnprocessableEntityException(
                    f"Field {f} is prohibited for updating",
                )

        result = await self.repository.update_by(
            field="id",
            value=id_,
            attributes=attributes,
            unique=True,
        )
        if result is None:
            raise NotFoundException(
                f"{self.model_class.__name__} with id: {id_} not found",
            )
        return result  # type: ignore[return-value]

    async def update_by_uuid(self, uuid: UUID, attributes: dict[str, Any]) -> ModelType:
        """Обновляет единственную сущность по её id."""
        for f in attributes:
            if f in self.exclude_fields:
                raise UnprocessableEntityException(
                    f"Field {f} is prohibited for updating",
                )

        result = await self.repository.update_by(
            field="id",
            value=uuid,
            attributes=attributes,
            unique=True,
        )
        if result is None:
            raise NotFoundException(
                f"{self.model_class.__name__} with uuid: {uuid} not found",
            )
        return result  # type: ignore[return-value]

    @transactional
    async def delete_by_id(self, id_: int) -> ModelType:
        """Удаляет единственную сущность по её id."""
        result = await self.repository.delete_by(field="id", value=id_, unique=True)
        if result is None:
            raise NotFoundException(
                f"{self.model_class.__name__} with id: {id_} not found",
            )
        return result  # type: ignore[return-value]

    @transactional
    async def delete_by_uuid(self, uuid: UUID) -> ModelType:
        """Удаляет единственную сущность по её id."""
        result = await self.repository.delete_by(field="id", value=uuid, unique=True)
        if result is None:
            raise NotFoundException(
                f"{self.model_class.__name__} with uuid: {uuid} not found",
            )
        return result  # type: ignore[return-value]

    @transactional
    async def create_or_update_by(
        self,
        attributes: dict[str, Any],
        update_fields: list[str] | None = None,
    ) -> ModelType:
        """Создать или обновить модель."""
        for f in attributes:
            if f in self.exclude_fields:
                raise UnprocessableEntityException(f"Field {f} is prohibited")
        result = await self.repository.create_or_update_by(
            attributes=attributes,
            update_fields=update_fields,
        )
        if result is None:
            raise NotFoundException("Failed to insert or update")
        return result

    async def make_pagination_response(
        self,
        data: Sequence,
        skip: int = 0,
        limit: int = 100,
        filter_request: FilterRequest | None = None,
    ) -> PaginationResponse:
        """Возвращает ответ для пагинации с численными полями.

        Args:
            data: Сущность или список сущностей ответа.
            skip: Кол-во записей для пропуска.
            limit: Кол-во записей для возврата.
            filter_request: Фильтры для совпадения.
        """
        total_count = (await self.count(filter_request=filter_request)).count
        page = skip // limit + 1 if limit > 0 else 1
        page_size = len(data)
        total_pages = (total_count + limit - 1) // limit if limit > 0 else 1

        return PaginationResponse(
            data=list(data),
            page=page,
            page_size=page_size,
            total_pages=total_pages,
            total_count=total_count,
        )

    @staticmethod
    async def extract_attributes_from_schema(
        schema: BaseModel,
        excludes: set[str | int] | None = None,
    ) -> dict[str, Any]:
        """Извлекает атрибуты из схемы.

        Args:
            schema: Схема для извлечения атрибутов.
            excludes: Атрибуты для исключения.

        Returns:
            Атрибуты.
        """
        return schema.model_dump(exclude=excludes, exclude_unset=True)

    def __repr__(self) -> str:
        """Возвращает строковое представление объекта.

        Returns:
            Строковое представление объекта.
        """
        return f"<{self.__class__.__name__}>"
